package com.bomostory.sceneeditmodule.utils

import android.content.ContentUris
import android.content.Context
import android.database.Cursor
import android.graphics.Bitmap
import android.net.Uri
import android.os.Build
import android.os.Environment
import android.provider.DocumentsContract
import android.provider.MediaStore
import com.bomostory.sceneeditmodule.Config
import com.bomostory.sceneeditmodule.SceneDrawer
import com.bomostory.sceneeditmodule.basicdata.Project
import com.google.gson.Gson
import io.reactivex.Completable
import java.io.*


object FileUtils {

    fun getProjectPath(project: Project): String {
        return File(Config.PROJECTS_FOLDER, project.name).absolutePath
    }

    fun getVideoFile(project: Project): File {
        val projectFolder = File(Config.PROJECTS_FOLDER, project.name)
        return File(projectFolder, Config.VIDEO_FILE_NAME)
    }

    fun getCoverFile(project: Project): File {
        val projectFolder = File(Config.PROJECTS_FOLDER, project.name)
        return File(projectFolder, Config.COVER_FILE_NAME)
    }

    fun getRealPathFromURI(context: Context, uri: Uri): String? {
        val isKitKat = Build.VERSION.SDK_INT >= 19

        if (isKitKat && DocumentsContract.isDocumentUri(context, uri)) {
            // DocumentProvider
            if (isExternalStorageDocument(uri)) {
                // ExternalStorageProvider
                val docId = DocumentsContract.getDocumentId(uri)
                val split = docId.split(":".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                val type = split[0]

                return if ("primary".equals(type, ignoreCase = true)) {
                    Environment.getExternalStorageDirectory().toString() + "/" + split[1]
                } else {
                    File(uri.toString()).absolutePath
                }
            } else if (isDownloadsDocument(uri)) {
                // DownloadsProvider
                val id = DocumentsContract.getDocumentId(uri)
                if (id.startsWith("raw:")) {
                    return id.replaceFirst("raw:", "");
                }
                try {
                    val contentUri = ContentUris.withAppendedId(Uri.parse("content://downloads/public_downloads"), id.toLong())
                    return getDataColumn(context, contentUri, null, null)
                } catch (e: NumberFormatException) {
                }

            } else if (isMediaDocument(uri)) {
                // MediaProvider
                val docId = DocumentsContract.getDocumentId(uri)
                val split = docId.split(":".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()
                val type = split[0]

                var contentUri: Uri? = null
                if ("image" == type) {
                    contentUri = MediaStore.Images.Media.EXTERNAL_CONTENT_URI
                } else if ("video" == type) {
                    contentUri = MediaStore.Video.Media.EXTERNAL_CONTENT_URI
                } else if ("audio" == type) {
                    contentUri = MediaStore.Audio.Media.EXTERNAL_CONTENT_URI
                }

                val selection = "_id=?"
                val selectionArgs = arrayOf(split[1])

                return getDataColumn(context, contentUri, selection, selectionArgs)
            }
        } else if ("content".equals(uri.scheme!!, ignoreCase = true)) {
            // Return the remote address
            if (isGooglePhotosUri(uri))
                return uri.lastPathSegment
            return getDataColumn(context, uri, null, null)
        } else if ("file".equals(uri.scheme!!, ignoreCase = true)) {
            // File
            return uri.path
        }
        return null
    }

    /**
     * Get the value of the data column for this Uri. This is useful for MediaStore Uris, and other file-based ContentProviders.
     *
     * @param context
     * The context.
     * @param uri
     * The Uri to query.
     * @param selection
     * (Optional) Filter used in the query.
     * @param selectionArgs
     * (Optional) Selection arguments used in the query.
     * @return The value of the _data column, which is typically a file path.
     */
    private fun getDataColumn(context: Context, uri: Uri?, selection: String?, selectionArgs: Array<String>?): String? {
        var cursor: Cursor? = null
        val column = "_data"
        val projection = arrayOf(column)

        try {
            cursor = context.contentResolver.query(uri!!, projection, selection, selectionArgs, null)
            if (cursor != null && cursor.moveToFirst()) {
                val index = cursor.getColumnIndexOrThrow(column)
                return cursor.getString(index)
            }
        } finally {
            cursor?.close()
        }
        return null
    }

    /**
     * @param uri
     * The Uri to check.
     * @return Whether the Uri authority is ExternalStorageProvider.
     */
    private fun isExternalStorageDocument(uri: Uri): Boolean {
        return "com.android.externalstorage.documents" == uri.authority
    }

    /**
     * @param uri
     * The Uri to check.
     * @return Whether the Uri authority is DownloadsProvider.
     */
    private fun isDownloadsDocument(uri: Uri): Boolean {
        return "com.android.providers.downloads.documents" == uri.authority
    }

    /**
     * @param uri
     * The Uri to check.
     * @return Whether the Uri authority is MediaProvider.
     */
    private fun isMediaDocument(uri: Uri): Boolean {
        return "com.android.providers.media.documents" == uri.authority
    }

    /**
     * @param uri
     * The Uri to check.
     * @return Whether the Uri authority is Google Photos.
     */
    private fun isGooglePhotosUri(uri: Uri): Boolean {
        return "com.google.android.apps.photos.content" == uri.authority
    }

    fun saveGifForActor(context: Context, project: Project, uri: Uri, fileName: String) {
        val dir = File("${getProjectPath(project)}/${Config.IMAGE_FOLDER_NAME}/")
        dir.mkdirs()
        val file = File(dir, fileName.plus(".gif"))
        try {
            var sourceFile = File(getRealPathFromURI(context, uri))
            val bis = BufferedInputStream(FileInputStream(sourceFile))
            val bos = BufferedOutputStream(FileOutputStream(file))
            val buf = ByteArray(1024)
            bis.read(buf)
            do {
                bos.write(buf)
            } while (bis.read(buf) !== -1)
            bos.flush()
            bos.close()
            bis.close()
            if (context is OnSaveActorImage) {
                context.onSaveActorImage(file.path)
            } else {
                throw RuntimeException(context.toString() + " must implement OnFragmentInteractionListener")
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    fun saveImageForActor(context: Context, project: Project, uri: Uri, fileName: String){
        val dir = File("${getProjectPath(project)}/${Config.IMAGE_FOLDER_NAME}/")
        dir.mkdirs()
        val file = File(dir, fileName.plus(".jpg"))
        try {
            val bitmap = MediaStore.Images.Media.getBitmap(context.contentResolver, uri)
            val bos = BufferedOutputStream(FileOutputStream(file))
            bitmap.compress(Bitmap.CompressFormat.JPEG, 80, bos)
            bos.flush()
            bos.close()
            if (context is OnSaveActorImage) {
                context.onSaveActorImage(file.path)
            } else {
                throw RuntimeException(context.toString() + " must implement OnFragmentInteractionListener")
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    fun saveBitmapForActor(context: Context, project: Project, bitmap: Bitmap?, fileName: String){
        val dir = File("${getProjectPath(project)}/${Config.IMAGE_FOLDER_NAME}/")
        dir.mkdirs()
        val file = File(dir, fileName.plus(".png"))
        try {
            val bos = BufferedOutputStream(FileOutputStream(file))
            bitmap?.compress(Bitmap.CompressFormat.PNG, 80, bos)
            bos.flush()
            bos.close()
            if (context is OnSaveActorImage) {
                context.onSaveActorImage(file.path)
            } else {
                throw RuntimeException(context.toString() + " must implement OnFragmentInteractionListener")
            }
        } catch (e: Exception) {
            e.printStackTrace()
        }
    }

    private fun readFileToString(file: File): String {
        if (!file.exists()) return ""
        return FileReader(file).run {
            val text = readText()
            close()
            text
        }
    }

    fun changeProjectName(oldName: String, newName: String): Project {
        val oldProjectFile = File("${Config.PROJECTS_FOLDER}/$oldName/${Config.PROJECT_FILE_NAME}")
        var oldProject = Gson().fromJson<Project>(readFileToString(oldProjectFile), Project::class.java)

        val oldProjectFolder = File(Config.PROJECTS_FOLDER, oldName)
        val newProjectFolder = File(Config.PROJECTS_FOLDER, newName)
        val success = oldProjectFolder.renameTo(newProjectFolder)

        if (!success) {
            return oldProject
        } else {
            val newProjectFile = File("${Config.PROJECTS_FOLDER}/$newName/${Config.PROJECT_FILE_NAME}")
            var oldProjectPath = File(Config.PROJECTS_FOLDER, oldName).absolutePath
            var newProjectPath = File(Config.PROJECTS_FOLDER, newName).absolutePath
            fun changeFilePath(file: File): File {
                println("resultProjectPath = ${file.absolutePath.replace(oldProjectPath, newProjectPath)}")
                return File(file.absolutePath.replace(oldProjectPath, newProjectPath))
            }
            println("oldProjectPath = $oldProjectPath")
            println("newProjectPath = $newProjectPath")
            val newProject = Project(
                    oldProject.themeAssetIndex,
                    changeFilePath(oldProject.projectFile),
                    changeFilePath(oldProject.projectFolder),
                    changeFilePath(oldProject.scene1File),
                    changeFilePath(oldProject.scene2File),
                    if (oldProject.coverFile != null) changeFilePath(oldProject.coverFile!!) else null,
                    changeFilePath(oldProject.assetFolder),
                    newName,
                    oldProject.author,
                    oldProject.frontCoverColor,
                    oldProject.backCoverColor,
                    oldProject.category,
                    oldProject.scene1,
                    oldProject.scene2,
                    oldProject.story,
                    oldProject.actorResource,
                    oldProject.sceneResource
            )

            val scenes = newProject.story?.scenes
            if (scenes != null) {
                var count = 0
                for (scene in scenes) {
                    for (layer in scene.layers) {
                        for (actor in layer.actors) {
                            actor.resourcePath = actor.resourcePath.replace(oldProjectPath, newProjectPath)
                            count++
                        }
                    }
                }
                println("count = $count")
            }
            return newProject

//                try {
//                    val fileWriter = FileWriter(newProjectFile)
//                    fileWriter.write(Gson().toJson(newProject))
//                    fileWriter.flush()
//                    fileWriter.close()
//                } catch (e: Exception) {
//                    e.printStackTrace()
//                }
        }
    }

    fun saveProject(context: Context, project: Project, scaleWidth: Int, scaleHeight: Int): Completable {
        return Completable.create {
            try {
                val fileWriter = FileWriter(File(File(Config.PROJECTS_FOLDER, project.name),Config.PROJECT_FILE_NAME))
                fileWriter.write(Gson().toJson(project))
                fileWriter.flush()
                when (project.story?.scenes?.size) {
                    1 -> {
                        var fileOutputStream = FileOutputStream(project.scene1File)
                        project.story?.let {
                            it.scenes?.let {
                                SceneDrawer.drawScene(context, it[0], 0, scaleWidth, scaleHeight)?.compress(Bitmap.CompressFormat.PNG,
                                        100, fileOutputStream)
                                fileOutputStream.flush()
                                fileOutputStream.close()
                            }
                        }
                    }
                    else -> {
                        var fileOutputStream = FileOutputStream(project.scene1File)
                        project.story?.let {
                            it.scenes?.let {
                                SceneDrawer.drawScene(context, it[0], 0, scaleWidth, scaleHeight)?.compress(Bitmap.CompressFormat.PNG,
                                        100, fileOutputStream)
                                fileOutputStream.flush()
                                fileOutputStream.close()
                            }
                        }
                        fileOutputStream = FileOutputStream(project.scene2File)
                        project.story?.let {
                            it.scenes?.let {
                                SceneDrawer.drawScene(context, it[1], 0, scaleWidth, scaleHeight)?.compress(Bitmap.CompressFormat.PNG,
                                        100, fileOutputStream)
                                fileOutputStream.flush()
                                fileOutputStream.close()
                            }
                        }
                    }
                }
                fileWriter.close()
            } catch (e: Exception) {
                e.printStackTrace()
            }
            it.onComplete()
        }
    }

    interface OnSaveActorImage {
        fun onSaveActorImage(filePath: String)
    }
}