package com.bomostory.sceneeditmodule.utils

import android.media.MediaPlayer
import com.bomostory.sceneeditmodule.basicdata.Music
import com.bomostory.sceneeditmodule.basicdata.Scene
import com.bomostory.sceneeditmodule.basicdata.Story
import java.util.*
import kotlin.collections.ArrayList
import kotlin.collections.HashMap

class MoviePlayerV2 {
    interface OnMoviePlayListener {
        fun onMovieSceneUpdate(scene: Scene)
        fun onMoviePlayViewUpdate(x: Int)
        fun onMoviePlayTimeUpdate(time: Long)
        fun onMoviePlayComplete()
    }

    interface OnGifPlayListener {
        fun onGifStart()
        fun onGifSeek(time: Int)
        fun onGifPause()
        fun onGifUpdate(time: Int)
    }

    var moviePlayListener: OnMoviePlayListener? = null
    var gifPlayListener: OnGifPlayListener? = null

    var story: Story? = null

    var period: Long = 0

    var timer: Timer? = null

    var currentTime: Long = 0

    var isPause = false

    var musicList: List<Music> = ArrayList()

    private var mediaPlayers = ArrayList<MediaPlayer>()

    private val recordPlayers = HashMap<Int, MediaPlayer>()


    fun init() {
        isPause = false

        timer?.cancel()
        timer?.purge()
        timer = null

        currentTime = 0

        updateMovie(true)

        tryStopRecord()

        tryStopMusic()

        initRecord()

        initMediaPlayer()

        gifPlayListener?.onGifPause()
    }

    fun start() {
        if (timer == null) {
            var time = System.nanoTime()
            timer = Timer()
            timer?.scheduleAtFixedRate(object : TimerTask() {
                override fun run() {
                    var now = System.nanoTime()
                    if (!isPause) {

                        currentTime = Math.min(period, currentTime + (now - time) / 1000000)

                        updateMovie()

                        tryPlayRecord()

                        tryPlayMusic()

                        if (currentTime >= period) {
                            moviePlayListener?.onMoviePlayComplete()
                            init()
                        }
                    }
                    time = now
                }
            }, 0, 20)
        } else {
            isPause = false
            updateMovie(true)
        }
        gifPlayListener?.onGifStart()
    }

    fun pause() {
        isPause = true

        tryPauseRecord()

        tryPauseMusic()

        gifPlayListener?.onGifPause()
    }

    fun seekTo(currentTime: Long) {
        this.currentTime = currentTime

        updateMovie(true)

        trySeekRecord()

        trySeekMusic()
    }

    private var startPeriodCache = 0L
    private var endPeriodCache = 0L
    private var targetScene: Scene? = null
    private var previousTrackIndex = -1
    private var currentTrackIndex = -1

    private fun updateMovie(seek: Boolean = false) {
        if (currentTime < startPeriodCache || currentTime >= endPeriodCache) {
            targetScene = null
            var startPeriod = 0L
            var endPeriod = 0L
            story?.apply {
                for (scene in scenes) {
                    scene.record?.apply {
                        endPeriod += period
                        if (currentTime in (endPeriod - period)..endPeriod) {
                            targetScene = scene
                            startPeriod = endPeriod - period
                        }
                    }
                    if (targetScene != null)
                        break
                }
            }
            startPeriodCache = startPeriod
            endPeriodCache = endPeriod
            previousTrackIndex = -1
            currentTrackIndex = -1
        }

        val currentTimeInScene = currentTime - startPeriodCache
        targetScene?.apply {
            moviePlayListener?.onMovieSceneUpdate(this)
            record?.let {
                if (it.tracks == null || it.tracks.size == 0) {
                    moviePlayListener?.onMoviePlayViewUpdate(0)
                    return@let
                }
                if (currentTimeInScene < it.tracks[0].time) {
                    moviePlayListener?.onMoviePlayViewUpdate(0)
                    return@let
                }
                if (currentTimeInScene >= it.tracks[it.tracks.size - 1].time) {
                    moviePlayListener?.onMoviePlayViewUpdate(it.tracks[it.tracks.size - 1].positionX)
                    return@let
                }

                var startIndex = 1
                var endIndex = it.tracks.size
                var hitCache = true
                if (currentTrackIndex > 0) {
                    if (currentTimeInScene < it.tracks[previousTrackIndex].time) {
                        endIndex = previousTrackIndex
                        hitCache = false
                    } else if (currentTimeInScene > it.tracks[currentTrackIndex].time) {
                        startIndex = Math.max(1, currentTrackIndex - 1)
                        hitCache = false
                    }
                } else {
                    hitCache = false
                }
                if (!hitCache) {
                    previousTrackIndex = Math.max(0, startIndex)
                    currentTrackIndex = Math.min(it.tracks.size - 1, endIndex)
                    for (i in startIndex until endIndex) {
                        if (currentTimeInScene in (it.tracks[i - 1].time .. it.tracks[i].time)) {
                            previousTrackIndex = i - 1
                            currentTrackIndex = i
                            break
                        }
                    }
                }

                val previousTrack = it.tracks[previousTrackIndex]
                val track = it.tracks[currentTrackIndex]
                val positionX = previousTrack.positionX + (track.positionX - previousTrack.positionX) * (currentTimeInScene - previousTrack.time) / Math.max(1, track.time - previousTrack.time)
                moviePlayListener?.onMoviePlayViewUpdate(positionX.toInt())
            }
        }
        moviePlayListener?.onMoviePlayTimeUpdate(currentTime)
        if (seek) {
            if (timer == null || isPause) {
                gifPlayListener?.onGifUpdate(currentTimeInScene.toInt())
                gifPlayListener?.onGifPause()
            }
            else {
                gifPlayListener?.onGifSeek(currentTimeInScene.toInt())
            }
        }
    }

    private fun getScene(t: Long): Scene? {
        var currentScene: Scene? = null
        var startPeriod = 0L
        var endPeriod = 0L

        story?.apply {
            for (scene in scenes) {
                scene.record?.apply {
                    endPeriod += period
                    if (t in startPeriod..endPeriod) {
                        currentScene = scene
                        startPeriod = endPeriod
                    }
                }
            }
        }
        return currentScene
    }

    private fun initRecord() {
        recordPlayers.clear()
        story?.let { story ->
            for (scene in story.scenes) {
                scene.recordPath?.let {
                    val mediaPlayer = MediaPlayer()
                    mediaPlayer.apply {
                        reset()
                        setDataSource(scene.recordPath)
                        prepare()
                    }
                    recordPlayers.put(story.scenes.indexOf(scene), mediaPlayer)
                }
            }
        }
    }

    private fun initMediaPlayer() {
        mediaPlayers.clear()

        for (music in musicList) {
            val mediaPlayer = MediaPlayer()
            mediaPlayer.apply {
                reset()
                setDataSource(music.path)
                isLooping = music.isLoop
                prepare()
            }
            mediaPlayers.add(mediaPlayer)
        }
    }

    private fun tryPlayRecord() {
        story?.let {
            recordPlayers[it.scenes.indexOf(getScene(currentTime))]?.apply {
                if (!isPlaying) {
                    start()
                }
            }
        }
    }

    private fun tryPlayMusic() {
        for (music in musicList) {
            val musicEndTime = if (music.isLoop) (music.offsetTime + music.endTime - music.startTime) + music.loopDuration else (music.offsetTime + music.endTime - music.startTime)

            if (currentTime in music.offsetTime..musicEndTime) {
                if (!mediaPlayers[musicList.indexOf(music)].isPlaying) {
                    mediaPlayers[musicList.indexOf(music)].start()
                } else if (currentTime >= musicEndTime) {
                    mediaPlayers[musicList.indexOf(music)].stop()
                    mediaPlayers[musicList.indexOf(music)].prepare()
                }
            }
        }
    }

    private fun tryPauseRecord() {
        recordPlayers.forEach {
            it.value.pause()
        }
    }

    private fun tryPauseMusic() {
        for (mediaPlayer in mediaPlayers) {
            mediaPlayer.pause()
        }
    }

    private fun trySeekRecord() {
        story?.let {
            var period: Long = 0
            for (i in 0 until it.scenes.indexOf(getScene(currentTime))) {
                period += (it.scenes[i].record?.period ?: 0)
            }

            recordPlayers[it.scenes.indexOf(getScene(currentTime))]?.apply {
                if (period == 0L) {
                    this.seekTo(currentTime.toInt())
                } else {
                    this.seekTo((currentTime % period).toInt())
                }
            }

        }
    }

    private fun trySeekMusic() {
        for (music in musicList) {
            val musicEndTime = if (music.isLoop) (music.offsetTime + music.endTime - music.startTime) + music.loopDuration else (music.offsetTime + music.endTime - music.startTime)

            if (currentTime in music.offsetTime..musicEndTime) {
                val actualTime = if (currentTime - music.offsetTime >= music.endTime - music.startTime) ((currentTime - music.offsetTime).toDouble() / (music.endTime - music.startTime)).toInt()
                else (currentTime - music.offsetTime).toInt()
                mediaPlayers[musicList.indexOf(music)].seekTo(actualTime)
            } else {
                mediaPlayers[musicList.indexOf(music)].stop()
                mediaPlayers[musicList.indexOf(music)].prepare()
            }
        }
    }

    private fun tryStopRecord() {
        recordPlayers.forEach {
            it.value.stop()
        }
    }

    private fun tryStopMusic() {
        for (mediaPlayer in mediaPlayers) {
            mediaPlayer.stop()
        }
    }
}